<?php
/**
 * @file
 * Contains \Drupal\inspect\Inspect.
 */

namespace Drupal\inspect;

use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;

use \SimpleComplex\Inspect\InspectorInterface as ParentInspectorInterface;

/**
 * Variable analyzer and exception tracer.
 *
 * Proxy class for Inspector.
 */
class Inspect extends \SimpleComplex\Inspect\Inspect {

  /**
   * @see \SimpleComplex\Inspect\Helper\Config
   * @see \Drupal\inspect\InspectHelperConfig
   */
  const CLASS_CONFIG = InspectHelperConfig::class;

  /**
   * @see \SimpleComplex\Inspect\Helper\Unicode
   * @see \Drupal\inspect\InspectHelperUnicode
   */
  const CLASS_UNICODE = InspectHelperUnicode::class;

  /**
   * @see \SimpleComplex\Inspect\Inspector
   * @see \Drupal\inspect\Inspector
   *
   * @var string
   */
  const CLASS_INSPECTOR = Inspector::class;

  /**
   * @see Inspect::getLoggerFactory()
   *
   * @var \Drupal\Core\Logger\LoggerChannelFactoryInterface
   */
  protected $loggerFactory;

  /**
   * @param ConfigFactoryInterface $config_factory
   * @param LoggerChannelFactoryInterface $logger_factory
   */
  public function __construct(ConfigFactoryInterface $config_factory, LoggerChannelFactoryInterface $logger_factory)
  {
    parent::__construct();

    $this->configure($config_factory->get('inspect.settings'));

    $this->loggerFactory = $logger_factory;
  }

  /**
   * {@inheritDoc}
   *
   * Real Drupal app root, like Symfony project directory; that is without
   * trailing /web.
   *
   * @see \Drupal::root()
   *
   * Issues concerning directory structure:
   * @see https://www.drupal.org/project/drupal/issues/1672986
   * @see https://www.drupal.org/project/drupal/issues/2385395
   */
  public function rootDir() : string
  {
    if (!$this->rootDirLength) {
      $this->rootDir = \Drupal::root();
      if (mb_stripos($this->rootDir, DIRECTORY_SEPARATOR . 'web', -4) !== FALSE
        && file_exists($this->rootDir . DIRECTORY_SEPARATOR . '..' . DIRECTORY_SEPARATOR . 'composer.json')
      ) {
        $this->rootDir = mb_substr($this->rootDir, 0, -4);
      }
      $this->rootDirLength = mb_strlen($this->rootDir);
      if (!$this->rootDirLength) {
        // Flag that root dir cannot be established.
        $this->rootDir = '';
        $this->rootDirLength = -1;
      }
    }
    return $this->rootDir;
  }

  /**
   * For the Inspector.
   *
   * Simpler than overriding Inspector constructor and every single Inspector
   * inspection/trace method.
   *
   * The logger channel concept is challenging in the sense that you cannot
   * simply inject a PSR logger.
   *
   * @see Inspector::log()
   *
   * @return LoggerChannelFactoryInterface
   */
  public function getLoggerFactory() : LoggerChannelFactoryInterface
  {
    return $this->loggerFactory;
  }


  // Deprecated methods to be removed shortly.----------------------------------

  /**
   * @deprecated  Will be removed shortly.
   *
   * @param mixed $var
   * @param mixed $options
   *    Ignored, may no longer be appliccable.
   *
   * @return boolean|NULL
   *   NULL: user isnt permitted to log inspections.
   *   FALSE: on error.
   */
  public static function log($var, $options = NULL) {
    @trigger_error(
      __CLASS__ . '::' . __METHOD__ . ' method is deprecated since version 8.x-3.0 and will be removed soon. Use '
      . '\Drupal::service(\'inspect.inspect\')->variable($subject)->log()' . ' instead.',
      E_USER_DEPRECATED
    );
    // Assume that permission still exists in system.
    if (!\Drupal::currentUser()->hasPermission('inspect log')) {
      return NULL;
    }
    try {
      \Drupal::service('inspect.inspect')->variable($var)->log();
      return TRUE;
    }
    catch (\Throwable $ignore) {}
    return FALSE;
  }

  /**
   * @deprecated  Defunct and will be removed shortly.
   *
   * @return boolean
   *   FALSE: always, filing no longer supported.
   */
  public static function file() {
    @trigger_error(
      __CLASS__ . '::' . __METHOD__
      . ' method is defunct and deprecated since version 8.x-3.0 and will be removed soon.',
      E_USER_DEPRECATED
    );
    // No do.
    return FALSE;
  }

  /**
   * @deprecated  Defunct and will be removed shortly.
   *
   * @return string
   *   Empty: always, getting no longer supported.
   */
  public static function get() {
    @trigger_error(
      __CLASS__ . '::' . __METHOD__
      . ' method is defunct and deprecated since version 8.x-3.0 and will be removed soon.',
      E_USER_DEPRECATED
    );
    // No do.
    return '';
  }

  /**
   * Calling this method statically is deprecated since version 8.x-3.0.
   *
   * {@inheritDoc}
   */
  public function trace(/*?\Throwable*/ $throwableOrNull, $options = []) : ParentInspectorInterface {
    if (!isset($this)) {
      @trigger_error(
        'Calling '
        . __CLASS__ . '::' . __METHOD__ . ' method statically is defunct and deprecated since version 8.x-3.0. Use '
        . '\Drupal::service(\'inspect.inspect\')->trace($throwableOrNull)->log()' . ' instead.',
        E_USER_DEPRECATED
      );
      // Assume that permission still exists in system.
      if (\Drupal::currentUser()->hasPermission('inspect log')) {
        \Drupal::service('inspect.inspect')->trace($throwableOrNull, ['wrappers' => 1])->log();
      }
      // We have to return (dummy) Inspector as promised by return type.
      return \Drupal::service('inspect.inspect')->variable(NULL);
    }

    return parent::trace($throwableOrNull, $options);
  }

  /**
   * @deprecated  Will be removed shortly.
   *
   * @param mixed $var
   * @param mixed $options
   *    Ignored, may no longer be appliccable.
   *
   * @return boolean|NULL
   *   NULL: user isnt permitted to log inspections.
   *   FALSE: on error.
   */
  public static function traceLog($var, $options = NULL) {
    @trigger_error(
      __CLASS__ . '::' . __METHOD__ . ' method is deprecated since version 8.x-3.0 and will be removed soon. Use '
      . '\Drupal::service(\'inspect.inspect\')->trace($throwableOrNull)->log()' . ' instead.',
      E_USER_DEPRECATED
    );
    // Assume that permission still exists in system.
    if (!\Drupal::currentUser()->hasPermission('inspect log')) {
      return NULL;
    }
    try {
      \Drupal::service('inspect.inspect')->trace($var)->log();
      return TRUE;
    }
    catch (\Throwable $ignore) {}
    return FALSE;
  }

  /**
   * @deprecated  Defunct and will be removed shortly.
   *
   * @return boolean
   *   FALSE: always, filing no longer supported.
   */
  public static function traceFile() {
    @trigger_error(
      __CLASS__ . '::' . __METHOD__
      . ' method is defunct and deprecated since version 8.x-3.0 and will be removed soon.',
      E_USER_DEPRECATED
    );
    // No do.
    return FALSE;
  }

  /**
   * @deprecated  Defunct and will be removed shortly.
   *
   * @return string
   *   Empty: always, getting no longer supported.
   */
  public static function traceGet() {
    @trigger_error(
      __CLASS__ . '::' . __METHOD__
      . ' method is defunct and deprecated since version 8.x-3.0 and will be removed soon.',
      E_USER_DEPRECATED
    );
    // No do.
    return '';
  }

  /**
   * @deprecated  Will be removed shortly.
   *
   * @param string $message
   * @param string $type
   * @param integer|string $severity
   *
   * @return boolean|NULL
   *   NULL: user isnt permitted to log inspections.
   *   FALSE: always, method deprecated.
   */
  public static function logMessage($message = '', $type = 'inspect', $severity = 'debug') {
    @trigger_error(
      __CLASS__ . '::' . __METHOD__ . ' method is deprecated since version 8.x-3.0 and will be removed soon. Use '
      . '\Drupal::logger(\'channel\')->log()' . ' instead.',
      E_USER_DEPRECATED
    );
    // Assume that permission still exists in system.
    if (!\Drupal::currentUser()->hasPermission('inspect log')) {
      return NULL;
    }
    \Drupal::logger($type)->log(
      $severity,
      \Drupal\Component\Utility\Unicode::truncateBytes($message, Inspector::OUTPUT_DEFAULT)
    );
    return FALSE;
  }

  /**
   * @deprecated  Defunct and will be removed shortly.
   *
   * @return boolean|NULL
   *   FALSE: always, filing no longer supported.
   */
  public static function fileMessage() {
    @trigger_error(
      __CLASS__ . '::' . __METHOD__
      . ' method is defunct and deprecated since version 8.x-3.0 and will be removed soon.',
      E_USER_DEPRECATED
    );
    // No do.
    return FALSE;
  }

  /**
   * @deprecated  Defunct and will be removed shortly.
   *
   * @return void
   *   Sends 403 header and exits.
   */
  public static function logFromFrontend() {
    @trigger_error(
      __CLASS__ . '::' . __METHOD__
      . ' method is defunct and deprecated since version 8.x-3.0 and will be removed soon.',
      E_USER_DEPRECATED
    );
    header('HTTP/1.1 403 Forbidden');
    exit;
  }

  /**
   * @deprecated  Defunct and will be removed shortly.
   *
   * @return integer
   *    Zero.
   */
  public static function outputMax() {
    @trigger_error(
      __CLASS__ . '::' . __METHOD__
      . ' method is defunct and deprecated since version 8.x-3.0 and will be removed soon.',
      E_USER_DEPRECATED
    );
    return 0;
  }

  /**
   * @deprecated  Defunct and will be removed shortly.
   *
   * @return string|boolean
   *   FALSE: always, feature no longer supported.
   */
  public static function ensureDirectory() {
    @trigger_error(
      __CLASS__ . '::' . __METHOD__
      . ' method is defunct and deprecated since version 8.x-3.0 and will be removed soon.',
      E_USER_DEPRECATED
    );
    return FALSE;
  }

  /**
   * @deprecated  Method no longer relevant in this context.
   *
   * Convert number or hex string to any base through 2 and 62.
   *
   * Base 62 produces up to 35% shorter string than base 16.
   *
   *  Like native base_convert() precision fails for numbers (somewhere) larger
   *  than (tested on windows box with ini precision 14):
   *  - base 10: 9007199254740991 (> 15 digits)
   *  - base 16: 1fffffffffffff (> 14 digits)
   *
   * @param integer|float|string $num
   * @param integer $fromBase
   *   Default: 10
   *   Only supports 10 and 16; if other, returns arg $num unaltered.
   * @param integer $toBase
   *   Default (and max): 62.
   *
   * @return string
   */
  public static function baseConvert($num, $fromBase = 10, $toBase = 62) {
    @trigger_error(
      __CLASS__ . '::' . __METHOD__ . ' method is deprecated since version 8.x-3.0 and will be removed soon.',
      E_USER_DEPRECATED
    );
    static $table = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
    $zero = $table[0];
    $remainder = $num;
    if ($num < 1) {
      if (!$num) {
        return '' . $zero; // zero
      }
      $remainder *= -1;
    }
    if ($toBase == $fromBase) {
      return $remainder;
    }
    switch ($fromBase) {
      case 10:
        break;
      case 16:
        $remainder = base_convert($remainder, 16, 10);
        break;
      default:
        return $remainder;
    }
    // Arg $num may be string, and base_convert() returns string.
    $remainder = (float)$remainder;
    if ($toBase > 62) {
      $toBase = 62;
    }
    elseif ($toBase < 2) {
      $toBase = 2;
    }
    $res = '';
    // Find power ~ required number of digits.
    $pow = 0;
    $divisor = 1;
    // Limit:100 probably way too high, havent seen more than 10 iterations
    // (1fffffffffffff, with precision 14)
    $limit = 100;
    while ($remainder >= ($test = pow($toBase, ++$pow)) && (--$limit)) {
      $divisor = $test;
      --$limit;
    }
    --$pow;
    // Do modulo procedure.
    // Limit:100 probably way too high, havent seen more than 14 iterations
    // (1fffffffffffff, with precision 14).
    $limit = 100;
    while ($remainder > 0 && (--$limit)) {
      if ($pow) {
        if ($divisor == $remainder) {
          return $res . $table[1] . str_repeat($zero, $pow);
        }
        elseif ($divisor > $remainder) {
          $res .= $zero;
        }
        else {
          if ($remainder > $divisor) {
            $res .= $table[ (int)floor($remainder / $divisor) ];
            $remainder = fmod($remainder, $divisor);
          }
          else {
            $res .= $table[ $pow ];
            $remainder = fmod($remainder, $divisor);
          }
          if ($remainder < 1 && !floor($remainder)) {
            return $res . str_repeat($zero, $pow);
          }
        }
      }
      else {
        return $res . $table[ (int)$remainder ];
      }
      $divisor = pow($toBase, --$pow);
    }
    return $res;
  }
}
